/*	XAppMenu.h
 *
 *		Define various menu constants for those menu items that
 *	are expected by the XGAppCore code
 */

/**********************************************************************

  YAAF is Copyright 1997-1999 by William Woody.

  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions 
  are met:

  Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.

  Neither the name of the copyright holder or other contributors may
  be used to promote products derived from this software without
  specific written prior permission.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS 
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN 
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
  SOFTWARE.
  
  Contact:   woody@alumni.caltech.edu
  Web page:  http://www.pandawave.com/yaaf/

**********************************************************************/

#ifndef PandaWave_XAPPMENU
#define PandaWave_XAPPMENU

#include <XConfig.h>

/**@name Support--Application Menu Command IDs
 *
 *		These are the command identifiers used by YAAF for various built-in
 *	menus. These are defined in <XAppMenu.h>.
 */

//@{

/************************************************************************/
/*																		*/
/*	Default Menu														*/
/*																		*/
/************************************************************************/

/// Default menu bar resource used by the XGApplication core when no menu is specified
#define	KXGDefaultMenu				128

/************************************************************************/
/*																		*/
/*	Menu Constants														*/
/*																		*/
/************************************************************************/

/// Command for 'Application:About Me' command
#define KXGAppAboutMe				1			// about application

/// Command for 'File:Close Window' command
#define KXGFileClose				2			// close front window

/// Command for 'File:Quit' command
#define KXGFileQuit					3			// quit application


/// Command for 'File:Save' command
#define KXGFileSave					13

/// Command for 'File:Save As' command
#define KXGFileSaveAs				14

/// Command for 'File:Revert' command
#define KXGFileRevert				15

/// Command for 'File:Save A Copy As' command
#define KXGFileSaveCopyAs			16


/// Command for 'Edit:Undo' command
#define KXGEditUndo					4			// standard edit menu

/// Command for 'Edit:Cut' command
#define KXGEditCut					5

/// Command for 'Edit:Copy' command
#define KXGEditCopy					6

/// Command for 'Edit:Paste' command
#define KXGEditPaste				7

/// Command for 'Edit:Clear' command
#define KXGEditClear				8

/// Command for 'Edit:Select All' command
#define KXGEditSelectAll			16


/// Command for 'Window:Tile Windows' command
#define KXGWindowTile				9

/// Command for 'Window:Cascade Windows' command
#define KXGWindowCascade			10

/// Command for 'Window:Arrange Windows' command
#define KXGWindowArrange			11

/// Command placeholder for first item in Window menu's window list.
#define KXGWindowList				12			// placeholder

//@}

#endif // PandaWave_XAPPMENU

